from laktory import models


pl = models.resources.databricks.DLTPipeline(
    name="pl-stock-prices",
    catalog="dev1",
    target="markets1",
)


def test_pipeline():
    print(pl.model_dump())
    assert pl.model_dump() == {
        "access_controls": [],
        "allow_duplicate_names": None,
        "catalog": "dev1",
        "channel": "PREVIEW",
        "clusters": [],
        "configuration": {},
        "continuous": None,
        "development": None,
        "edition": None,
        "libraries": None,
        "name": "pl-stock-prices",
        "notifications": [],
        "photon": None,
        "serverless": None,
        "storage": None,
        "target": "markets1",
    }


def test_pipeline_pulumi():
    print(pl.resource_name)
    assert pl.resource_name == "dlt-pl-stock-prices"
    assert pl.options.model_dump(exclude_none=True) == {
        "depends_on": [],
        "delete_before_replace": True,
    }
    print(pl.pulumi_properties)
    assert pl.pulumi_properties == {
        "catalog": "dev1",
        "channel": "PREVIEW",
        "clusters": [],
        "configuration": {},
        "name": "pl-stock-prices",
        "notifications": [],
        "target": "markets1",
    }

    # Resources
    assert len(pl.core_resources) == 1
    r = pl.core_resources[-1]
    r.options.aliases = ["my-file"]
    assert pl.core_resources[-1].options.aliases == ["my-file"]


if __name__ == "__main__":
    test_pipeline()
    test_pipeline_pulumi()



import os
import shutil
import pytest
from pyspark.errors import AnalysisException
import polars as pl

from laktory.models import TableDataSink
from laktory.models import FileDataSink
from laktory._testing import df_slv
from laktory._testing import df_slv_polars
from laktory._testing import Paths
from laktory._testing import spark

paths = Paths(__file__)


def test_file_data_sink():

    dirpath = os.path.join(paths.tmp, "df_slv_sink")
    if os.path.exists(dirpath):
        shutil.rmtree(dirpath)

    # Write as overwrite
    sink = FileDataSink(
        path=dirpath,
        format="PARQUET",
        mode="OVERWRITE",
    )
    sink.write(df_slv)

    # Write as append
    sink.write(df_slv, mode="append")

    # Write and raise error
    with pytest.raises(AnalysisException):
        sink.write(df_slv, mode="error")

    # Read back
    df = sink.as_source().read(spark=spark)

    # Test
    assert df.count() == 2 * df_slv.count()

    # Cleanup
    if os.path.exists(dirpath):
        shutil.rmtree(dirpath)


def test_file_data_sink_polars_parquet():

    filepath = os.path.join(paths.tmp, "df_slv_polars_sink.parquet")

    if os.path.exists(filepath):
        os.remove(filepath)

    # Write parquet
    sink = FileDataSink(
        path=filepath,
        format="PARQUET",
    )
    sink.write(df_slv_polars)

    # Write as append
    with pytest.raises(ValueError):
        sink.write(df_slv_polars, mode="append")

    # Read back
    source = sink.as_source()
    source.dataframe_type = "POLARS"
    df = source.read(filepath)

    # Test
    assert df.height == df_slv.count()
    assert df.columns == df_slv.columns

    # Cleanup
    if os.path.exists(filepath):
        os.remove(filepath)


def test_file_data_sink_polars_delta():

    dirpath = os.path.join(paths.tmp, "df_slv_polars_sink.delta")
    if os.path.exists(dirpath):
        shutil.rmtree(dirpath)

    # Write as overwrite
    sink = FileDataSink(
        path=dirpath,
        format="DELTA",
        mode="OVERWRITE",
    )
    sink.write(df_slv_polars)

    # Write as append
    sink.write(df_slv_polars, mode="append")

    # Read back
    source = sink.as_source()
    source.dataframe_type = "POLARS"
    df = source.read()

    # Test
    assert df.height == df_slv.count() * 2
    assert df.columns == df_slv.columns

    # Cleanup
    if os.path.exists(dirpath):
        shutil.rmtree(dirpath)


def test_table_data_sink():

    # Write as overwrite
    sink = TableDataSink(
        catalog_name="hive_metastore",
        schema_name="default",
        table_name="slv_stock_prices_sink",
        mode="OVERWRITE",
    )

    assert sink.full_name == "hive_metastore.default.slv_stock_prices_sink"
    assert sink.format == "DELTA"
    assert sink.mode == "OVERWRITE"

    # TODO: Test write using spark sessions on Databricks
    # source.write(df_slv)


if __name__ == "__main__":
    test_file_data_sink()
    test_file_data_sink_polars_parquet()
    test_file_data_sink_polars_delta()
    test_table_data_sink()



from typing import Union
from typing import Any
from pydantic import Field

from laktory.models.datasources.basedatasource import BaseDataSource
from laktory.spark import SparkDataFrame
from laktory.polars import PolarsDataFrame
from laktory._logger import get_logger

logger = get_logger(__name__)


class PipelineNodeDataSource(BaseDataSource):
    """
    Data Source using an upstream pipeline node. Using a pipeline node data
    source defines the interdependencies between each node in a pipeline.
    Depending on the selected pipeline orchestrator and the context, a pipeline
    node data source might read the data from:

    - memory
    - upstream node sink
    - DLT table

    Attributes
    ----------
    node_name:
        Name of the upstream pipeline node

    Examples
    ---------
    ```python
    from laktory import models

    brz = models.PipelineNode(
        name="brz_stock_prices",
        source={"path": "/Volumes/sources/landing/events/yahoo-finance/stock_price"},
        sink={"path": "/Volumes/sources/landing/tables/brz_stock_prices"},
    )

    slv = models.PipelineNode(
        name="slv_stock_prices",
        source={"node_name": "brz_stock_prices"},
        sink={"path": "/Volumes/sources/landing/tables/slv_stock_prices"},
    )

    pl = models.Pipeline(name="pl-stock-prices", nodes=[brz, slv])

    # pl.execute(spark=spark)
    ```
    """

    node_name: Union[str, None]
    node: Any = Field(None, exclude=True)  # Add suggested type?
    # include_failed_expectations: bool = True  # TODO: Implement
    # include_passed_expectations: bool = True  # TODO: Implement

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def _id(self) -> str:
        return self.node_name

    # ----------------------------------------------------------------------- #
    # Readers                                                                 #
    # ----------------------------------------------------------------------- #

    def _read_spark(self, spark) -> SparkDataFrame:

        # Reading from DLT
        if self.is_orchestrator_dlt:

            from laktory.dlt import read as dlt_read
            from laktory.dlt import read_stream as dlt_read_stream
            from laktory.dlt import is_debug

            if is_debug():
                df = None
                if self.node.output_df:
                    logger.info(
                        f"Reading pipeline node {self._id} from output DataFrame (DLT debug)"
                    )
                    df = self.node.output_df
                elif self.node.sink:
                    logger.info(
                        f"Reading pipeline node {self._id} from sink (DLT debug)"
                    )
                    df = self.node.sink.read(spark=spark, as_stream=self.as_stream)
                else:
                    logger.info(f"Can't read pipeline node {self._id} (DLT DEBUG)")

            else:
                if self.as_stream:
                    logger.info(f"Reading pipeline node {self._id} with DLT as stream")
                    df = dlt_read_stream(self.node.name)
                else:
                    logger.info(f"Reading pipeline node {self._id} with DLT as static")
                    df = dlt_read(self.node.name)

        # Read from node output DataFrame (if available)
        elif self.node.output_df is not None:
            logger.info(f"Reading pipeline node {self._id} from output DataFrame")
            df = self.node.output_df

        # Read from node sink
        elif self.node.sink:
            logger.info(f"Reading pipeline node {self._id} from sink")
            df = self.node.sink.read(spark=spark, as_stream=self.as_stream)

        else:
            raise ValueError(f"Pipeline Node {self._id} can't read DataFrame")

        return df

    def _read_polars(self) -> PolarsDataFrame:

        # Read from node output DataFrame (if available)
        if self.node.output_df is not None:
            logger.info(f"Reading pipeline node {self._id} from output DataFrame")
            df = self.node.output_df

        # Read from node sink
        elif self.node.sink:
            logger.info(f"Reading pipeline node {self._id} from sink")
            df = self.node.sink.read(as_stream=self.as_stream)

        else:
            raise ValueError(f"Pipeline Node {self._id} can't read DataFrame")

        return df



import os
import yaml
from typing import Any
from typing import Union
from pydantic import Field

from laktory._useragent import set_databricks_sdk_upstream
from laktory._logger import get_logger
from laktory._settings import settings
from laktory.constants import CACHE_ROOT
from laktory.models.basemodel import BaseModel

logger = get_logger(__name__)


class ConfigValue(BaseModel):
    type: str = "String"
    description: str = None
    default: Any = None


class PulumiStack(BaseModel):
    """
    A Pulumi stack is pulumi-specific flavor of the `laktory.models.Stack`. It
    re-structure the attributes to be aligned with a Pulumi.yaml file.

    It is generally not instantiated directly, but rather created using
    `laktory.models.Stack.to_pulumi()`.

    References
    ----------
    - pulumi yaml [options](https://www.pulumi.com/docs/languages-sdks/yaml/yaml-language-reference/)
    """

    name: str
    organization: str = Field(None, exclude=True)
    runtime: str = "yaml"
    description: Union[str, None] = None
    config: dict[str, Union[str, ConfigValue]] = {}
    variables: dict[str, Any] = {}
    resources: dict[str, Any] = {}
    outputs: dict[str, str] = {}

    def model_dump(self, *args, **kwargs) -> dict[str, Any]:
        """Serialize model to match the structure of a Pulumi.yaml file."""
        settings.camel_serialization = True
        kwargs["exclude_none"] = kwargs.get("exclude_none", True)
        d = super().model_dump(*args, **kwargs)

        # Special treatment of resources
        for r in self.resources.values():
            d["resources"][r.resource_name] = {
                "type": r.pulumi_resource_type,
                "properties": r.pulumi_properties,
                "options": r.options.model_dump(exclude_none=True),
            }
        settings.camel_serialization = False

        # Pulumi YAML requires the keyword "resources." to be removed
        pattern = r"\$\{resources\.(.*?)\}"
        self.variables[pattern] = r"${\1}"
        d = self.inject_vars(d)
        del self.variables[pattern]

        return d

    # ----------------------------------------------------------------------- #
    # Pulumi Methods                                                          #
    # ----------------------------------------------------------------------- #

    def write(self) -> str:
        """
        Write Pulumi.yaml configuration file

        Returns
        -------
        :
            Filepath of the configuration file
        """
        filepath = os.path.join(CACHE_ROOT, "Pulumi.yaml")

        if not os.path.exists(CACHE_ROOT):
            os.makedirs(CACHE_ROOT)

        with open(filepath, "w") as fp:
            yaml.dump(self.model_dump(), fp)

        return filepath

    def _call(self, command: str, stack: str, flags: list[str] = None):
        from laktory.cli._common import Worker

        self.write()
        worker = Worker()

        cmd = ["pulumi", command]
        cmd += ["-s", stack]

        if flags is not None:
            cmd += flags

        # Inject user-agent value for monitoring usage as a Databricks partner
        set_databricks_sdk_upstream()

        worker.run(
            cmd=cmd,
            cwd=CACHE_ROOT,
            raise_exceptions=settings.cli_raise_external_exceptions,
        )

    def preview(self, stack: str = None, flags: list[str] = None) -> None:
        """
        Runs `pulumi preview`

        Parameters
        ----------
        stack:
            Name of the stack to use
        flags:
            List of flags / options for pulumi preview
        """
        self._call("preview", stack=stack, flags=flags)

    def up(self, stack: str = None, flags: list[str] = None):
        """
        Runs `pulumi up`

        Parameters
        ----------
        stack:
            Name of the stack to use
        flags:
            List of flags / options for pulumi up
        """
        self._call("up", stack=stack, flags=flags)



# MAGIC %pip install laktory

# COMMAND ----------
import importlib
import sys
import os
import pyspark.sql.functions as F

from laktory import dlt
from laktory import models
from laktory import get_logger
from laktory import settings

dlt.spark = spark
logger = get_logger(__name__)

# Read pipeline definition
pl_name = spark.conf.get("pipeline_name", "dlt-stock-prices")
filepath = f"/Workspace{settings.workspace_laktory_root}pipelines/{pl_name}.json"
with open(filepath, "r") as fp:
    pl = models.Pipeline.model_validate_json(fp.read())


# Import User Defined Functions
sys.path.append(f"/Workspace{settings.workspace_laktory_root}pipelines/")
udfs = []
for udf in pl.udfs:
    if udf.module_path:
        sys.path.append(os.path.abspath(udf.module_path))
    module = importlib.import_module(udf.module_name)
    udfs += [getattr(module, udf.function_name)]

# --------------------------------------------------------------------------- #
# Tables and Views Definition                                                 #
# --------------------------------------------------------------------------- #


def define_table(node):
    @dlt.table_or_view(
        name=node.name,
        comment=node.description,
        as_view=node.sink is None,
    )
    @dlt.expect_all(node.warning_expectations)
    @dlt.expect_all_or_drop(node.drop_expectations)
    @dlt.expect_all_or_fail(node.fail_expectations)
    def get_df():
        logger.info(f"Building {node.name} node")

        # Execute node
        df = node.execute(spark=spark, udfs=udfs)
        df.printSchema()

        # Return
        return df

    return get_df


# --------------------------------------------------------------------------- #
# CDC tables                                                                  #
# --------------------------------------------------------------------------- #


def define_cdc_table(node):
    dlt.create_streaming_table(
        name=node.name,
        comment=node.comment,
    )

    df = dlt.apply_changes(**node.apply_changes_kwargs)

    return df


# --------------------------------------------------------------------------- #
# Execution                                                                   #
# --------------------------------------------------------------------------- #

# Build nodes
for node in pl.nodes:

    if node.dlt_template != "DEFAULT":
        continue

    if node.is_from_cdc:
        df = define_cdc_table(node)
        display(df)

    else:
        wrapper = define_table(node)
        df = dlt.get_df(wrapper)
        display(df)



from typing import Any
from typing import Union
from typing import Literal

from pydantic import model_validator

from laktory._logger import get_logger
from laktory._settings import settings
from laktory.models.basemodel import BaseModel
from laktory.models.resources.databricks.grants import Grants
from laktory.models.grants.tablegrant import TableGrant
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource
from laktory.models.resources.databricks.column import Column

logger = get_logger(__name__)


class Table(BaseModel, PulumiResource, TerraformResource):
    """
    A table resides in the third layer of Unity Catalog’s three-level
    namespace. It contains rows of data.

    Attributes
    ----------
    catalog_name:
        Name of the catalog storing the table
    columns:
        List of columns stored in the table
    comment:
        Text description of the catalog
    data:
        Data to be used to populate the rows
    grants:
        List of grants operating on the schema
    name:
        Name of the table
    primary_key:
        Name of the column storing a unique identifier for each row. It is used
        by the builder to drop duplicated rows.
    schema_name:
        Name of the schema storing the table
    table_type:
        Distinguishes a view vs. managed/external Table.
    view_definition:
        SQL text defining the view (for `table_type == "VIEW"`). Not supported
        for MANAGED or EXTERNAL table_type.
    warehouse_id:
        All table CRUD operations must be executed on a running cluster or SQL
        warehouse. If a warehouse_id is specified, that SQL warehouse will be
        used to execute SQL commands to manage this table.

    Examples
    --------
    ```py
    from laktory import models

    table = models.resources.databricks.Table(
        name="slv_stock_prices",
    )
    ```

    References
    ----------

    * [Databricks Unity Table](https://docs.databricks.com/en/data-governance/unity-catalog/index.html#tables)
    * [Pulumi Databricks Table](https://www.pulumi.com/registry/packages/databricks/api-docs/sqltable/)
    """

    catalog_name: Union[str, None] = None
    columns: list[Column] = []
    comment: Union[str, None] = None
    data_source_format: str = "DELTA"
    grants: list[TableGrant] = None
    name: str
    primary_key: Union[str, None] = None
    schema_name: Union[str, None] = None
    table_type: Literal["MANAGED", "EXTERNA", "VIEW"] = "MANAGED"
    view_definition: str = None
    warehouse_id: str = None

    # ----------------------------------------------------------------------- #
    # Validators                                                              #
    # ----------------------------------------------------------------------- #

    @model_validator(mode="after")
    def assign_catalog_schema(self) -> Any:
        # Assign to columns
        for c in self.columns:
            c.table_name = self.name
            c.catalog_name = self.catalog_name
            c.schema_name = self.schema_name

        # Warehouse ID
        if self.warehouse_id is None:
            self.warehouse_id = settings.databricks_warehouse_id

        return self

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def parent_full_name(self) -> str:
        """Schema full name `{catalog_name}.{schema_name}`"""
        _id = ""
        if self.catalog_name:
            _id += self.catalog_name

        if self.schema_name:
            if _id == "":
                _id = self.schema_name
            else:
                _id += f".{self.schema_name}"

        return _id

    @property
    def full_name(self) -> str:
        """Table full name `{catalog_name}.{schema_name}.{table_name}`"""
        _id = self.name
        if self.parent_full_name is not None:
            _id = f"{self.parent_full_name}.{_id}"
        return _id

    @property
    def column_names(self) -> list[str]:
        """List of column names"""
        return [c.name for c in self.columns]

    @property
    def is_from_cdc(self) -> bool:
        """If `True` CDC source is used to build the table"""
        return self.builder.is_from_cdc

    # ----------------------------------------------------------------------- #
    #  Methods                                                                #
    # ----------------------------------------------------------------------- #

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        """Table full name (catalog.schema.table)"""
        return self.full_name

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - table grants
        """
        resources = []

        # Schema grants
        if self.grants:
            resources += [
                Grants(
                    resource_name=f"grants-{self.resource_name}",
                    table=self.full_name,
                    grants=[
                        {"principal": g.principal, "privileges": g.privileges}
                        for g in self.grants
                    ],
                )
            ]
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:SqlTable"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return [
            "columns",
            "grants",
            "primary_key",
        ]

    @property
    def pulumi_properties(self):
        d = super().pulumi_properties
        d["columns"] = []
        for i, c in enumerate(self.columns):
            d["columns"] += [
                {
                    "name": c.name,
                    "comment": c.comment,
                    "type": c.type,
                }
            ]
        return d

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_sql_table"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes

    @property
    def terraform_properties(self) -> dict:
        d = super().terraform_properties
        d["column"] = []
        for i, c in enumerate(self.columns):
            d["column"] += [
                {
                    "name": c.name,
                    "comment": c.comment,
                    "type": c.type,
                }
            ]
        return d



import os
import shutil

from pyspark.sql import functions as F

from laktory import models
from laktory._testing import spark
from laktory._testing import Paths

paths = Paths(__file__)

# Data
df_brz = spark.read.parquet(os.path.join(paths.data, "./brz_stock_prices"))
df_slv = spark.read.parquet(os.path.join(paths.data, "./slv_stock_prices"))


def test_execute():

    sink_path = os.path.join(paths.tmp, "pl_node_sink")

    node = models.PipelineNode(
        name="slv_stock_prices",
        source={
            "table_name": "brz_stock_prices",
            "mock_df": df_brz,
        },
        transformer={
            "spark": True,
            "nodes": [
                {
                    "with_column": {
                        "name": "created_at",
                        "type": "timestamp",
                        "sql_expr": "data.created_at",
                    },
                },
                {
                    "with_column": {"name": "symbol", "sql_expr": "data.symbol"},
                },
                {
                    "with_column": {
                        "name": "close",
                        "type": "double",
                        "sql_expr": "data.close",
                    },
                },
                {
                    "func_name": "drop",
                    "func_args": ["data", "producer", "name", "description"],
                },
            ],
        },
        sink={
            "path": sink_path,
            "format": "PARQUET",
            "mode": "OVERWRITE",
        },
    )
    df0 = node.execute()
    df1 = spark.read.format("PARQUET").load(sink_path)

    assert df1.columns == df0.columns
    assert df1.columns == ["created_at", "symbol", "close"]
    assert df1.count() == df_brz.count()

    # Cleanup
    shutil.rmtree(sink_path)


def test_bronze():
    node = models.PipelineNode(
        layer="BRONZE",
        name="slv_stock_prices",
        source={
            "table_name": "brz_stock_prices",
            "mock_df": df_brz,
        },
        transformer={
            "nodes": [
                {
                    "with_column": {"name": "symbol", "sql_expr": "data.symbol"},
                },
            ]
        },
    )

    # Read and process
    df = node.execute(spark)

    # Test
    assert not node.drop_source_columns
    assert not node.drop_duplicates
    assert node.layer == "BRONZE"
    assert node.add_layer_columns
    assert df.columns == [
        "name",
        "description",
        "producer",
        "data",
        "symbol",
        "_bronze_at",
    ]
    assert df.count() == 80


def test_silver():
    df = df_brz.select(df_brz.columns)
    df = df.withColumn("_bronze_at", F.current_timestamp())
    df = df.union(df)

    node = models.PipelineNode(
        layer="SILVER",
        source={
            "table_name": "slv_stock_prices",
            "mock_df": df,
        },
        transformer={
            "nodes": [
                {
                    "with_column": {
                        "name": "symbol",
                        "type": "string",
                        "sql_expr": "data.symbol",
                    },
                },
            ]
        },
        drop_duplicates=True,
    )

    # Read and process
    df = node.execute(spark)

    # Test
    assert node.drop_source_columns
    assert node.drop_duplicates
    assert node.layer == "SILVER"
    assert node.add_layer_columns
    assert df.columns == ["_bronze_at", "symbol", "_silver_at"]
    assert df.count() == 80


def test_cdc():
    node = models.PipelineNode(
        source={
            "table_name": "brz_users_type1",
            "cdc": {
                "primary_keys": ["userId"],
                "sequence_by": "sequenceNum",
                "apply_as_deletes": "operation = 'DELETE'",
                "scd_type": 1,
                "except_columns": ["operation", "sequenceNum"],
            },
        },
    )

    # TODO: Test CDC transformations when ready
    print(node)


#
#     assert table.builder.apply_changes_kwargs == {
#         "apply_as_deletes": "operation = 'DELETE'",
#         "apply_as_truncates": None,
#         "column_list": [],
#         "except_column_list": ["operation", "sequenceNum"],
#         "ignore_null_updates": None,
#         "keys": ["userId"],
#         "sequence_by": "sequenceNum",
#         "source": "brz_users_cdc",
#         "stored_as_scd_type": 1,
#         "target": "brz_users_type1",
#         "track_history_column_list": None,
#         "track_history_except_column_list": None,
#     }
#     assert table.builder.is_from_cdc
#
#     # TODO: Run test with demo data
#     # from pyspark.sql import SparkSession
#
#     # spark = SparkSession.builder.appName("UnitTesting").getOrCreate()
#     #
#     # df_cdc = spark.createDataFrame(pd.DataFrame({
#     #     "userId": [124, 123, 125, 126, 123, 125, 125, 123],
#     #     "name": ["Raul", "Isabel", "Mercedes", "Lily", None, "Mercedes", "Mercedes", "Isabel"],
#     #     "city": ["Oaxaca", "Monterrey", "Tijuana", "Cancun", None, "Guadalajara", "Mexicali", "Chihuahua"],
#     #     "operation": ["INSERT", "INSERT", "INSERT", "INSERT", "DELETE", "UPDATE", "UPDATE", "UPDATE"],
#     #     "sequenceNum": [1, 1, 2, 2, 6, 6, 5, 5],
#     # }))
#     # df_cdc.show()


if __name__ == "__main__":
    test_execute()
    test_bronze()
    test_silver()
    test_cdc()



from typing import Literal

from laktory._logger import get_logger
from laktory.models.basemodel import BaseModel

logger = get_logger(__name__)


class PipelineNodeExpectation(BaseModel):
    """
    Definition of an expectation for a given node output DataFrame. It consists
    of an SQL expression of the expectation and of an action to be taken if the
    expectation is not met. Similar to DLT Table Expectation. Currently only
    supported with DLT pipeline orchestrator.

    Attributes
    ----------
    name:
        Name of the expectation
    expression:
        SQL expression definition the constraint
    action:
        Action to take when expectation is not met.
        `ALLOW`: Write invalid records to the output DataFrame, but log
        exception.
        `DROP`: Drop Invalid records to the output DataFrame and log exception.
        `FAIL`: Raise exception when invalid records are found.

    Examples
    --------
    ```py
    from laktory import models

    e = models.PipelineNodeExpectation(
        name="valid timestamp",
        expression="col(“timestamp”) > '2012-01-01'",
        action="DROP",
    )
    print(e)
    '''
    variables={} name='valid timestamp' expression="col(“timestamp”) > '2012-01-01'" action='DROP'
    '''
    ```

    References
    ----------

    * [DLT Table Expectations](https://docs.databricks.com/en/delta-live-tables/expectations.html)
    """

    name: str
    expression: str
    action: Literal["ALLOW", "DROP", "FAIL"] = "ALLOW"



import pytest
from pydantic import ValidationError

from laktory.models.resources.databricks import Table
from laktory._testing.stockprices import spark


def test_model():

    table = Table(
        name="slv_stock_prices",
        columns=[
            {
                "name": "created_at",
                "type": "timestamp",
            },
            {
                "name": "symbol",
                "type": "string",
            },
            {
                "name": "open",
                "type": "double",
            },
            {
                "name": "close",
                "type": "double",
            },
        ],
        catalog_name="dev",
        schema_name="markets",
    )

    data = table.model_dump()
    print(data)
    assert data == {
        "catalog_name": "dev",
        "columns": [
            {
                "catalog_name": "dev",
                "comment": None,
                "name": "created_at",
                "pii": None,
                "raise_missing_arg_exception": True,
                "schema_name": "markets",
                "table_name": "slv_stock_prices",
                "type": "timestamp",
                "unit": None,
            },
            {
                "catalog_name": "dev",
                "comment": None,
                "name": "symbol",
                "pii": None,
                "raise_missing_arg_exception": True,
                "schema_name": "markets",
                "table_name": "slv_stock_prices",
                "type": "string",
                "unit": None,
            },
            {
                "catalog_name": "dev",
                "comment": None,
                "name": "open",
                "pii": None,
                "raise_missing_arg_exception": True,
                "schema_name": "markets",
                "table_name": "slv_stock_prices",
                "type": "double",
                "unit": None,
            },
            {
                "catalog_name": "dev",
                "comment": None,
                "name": "close",
                "pii": None,
                "raise_missing_arg_exception": True,
                "schema_name": "markets",
                "table_name": "slv_stock_prices",
                "type": "double",
                "unit": None,
            },
        ],
        "comment": None,
        "data_source_format": "DELTA",
        "grants": None,
        "name": "slv_stock_prices",
        "primary_key": None,
        "schema_name": "markets",
        "table_type": "MANAGED",
        "view_definition": None,
        "warehouse_id": "08b717ce051a0261",
    }


if __name__ == "__main__":
    test_model()



from typing import Literal
from typing import Union
from laktory.models.datasinks.basedatasink import BaseDataSink
from laktory.spark import SparkDataFrame
from laktory.models.datasources.tabledatasource import TableDataSource
from laktory._logger import get_logger

logger = get_logger(__name__)


class TableDataSink(BaseDataSink):
    """
    Data Table data sink such as table on a Databricks catalog or on a
    data warehouse such as Snowflake, BigQuery, etc.

    Attributes
    ----------
    checkpoint_location:
        Path to which the checkpoint file for streaming dataframe should
        be written.
    catalog_name:
        Name of the catalog of the source table
    table_name:
        Name of the source table
    schema_name:
        Name of the schema of the source table
    warehouse:
        Type of warehouse to which the table should be published

    Examples
    ---------
    ```python
    from laktory import models
    import pandas as pd

    df = spark.createDataFrame(
        pd.DataFrame(
            {
                "symbol": ["AAPL", "GOOGL"],
                "price": [200.0, 205.0],
                "tstamp": ["2023-09-01", "2023-09-01"],
            }
        )
    )

    sink = models.TableDataSink(
        catalog_name="/Volumes/sources/landing/events/yahoo-finance/stock_price",
        schema_name="finance",
        table_name="slv_stock_prices",
        mode="OVERWRITE",
    )
    # sink.write(df)
    ```
    """

    catalog_name: Union[str, None] = None
    checkpoint_location: Union[str, None] = None
    format: Literal["DELTA", "PARQUET"] = "DELTA"
    schema_name: Union[str, None] = None
    table_name: Union[str, None]
    warehouse: Union[Literal["DATABRICKS"], None] = "DATABRICKS"

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def full_name(self) -> str:
        """Table full name {catalog_name}.{schema_name}.{table_name}"""
        if self.table_name is None:
            return None

        name = ""
        if self.catalog_name is not None:
            name = self.catalog_name

        if self.schema_name is not None:
            if name == "":
                name = self.schema_name
            else:
                name += f".{self.schema_name}"

        if name == "":
            name = self.table_name
        else:
            name += f".{self.table_name}"

        return name

    @property
    def _id(self) -> str:
        return self.full_name

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #

    def _write_spark(self, df: SparkDataFrame, mode=None) -> None:

        if mode is None:
            mode = self.mode

        if self.warehouse == "DATABRICKS":
            return self._write_spark_databricks(df, mode=mode)
        else:
            raise NotImplementedError(
                f"Warehouse '{self.warehouse}' is not yet supported."
            )

    def _write_spark_databricks(self, df: SparkDataFrame, mode) -> None:

        if df.isStreaming:
            logger.info(f"Writing {self._id} as stream with mode {self.mode}")
            writer = (
                df.writeStream.outputMode(self.mode)
                .format(self.format)
                .trigger(availableNow=True)  # TODO: Add option for trigger?
                .option("mergeSchema", "true")
            )

            if self.checkpoint_location:
                writer = writer.option("checkpointLocation", self.checkpoint_location)

            writer.toTable(self.full_name)

        else:
            logger.info(f"Writing {self._id} as static with mode {self.mode}")
            (
                df.write.format(self.format)
                .mode(mode)
                .option("mergeSchema", "true")
                .saveAsTable(self.full_name)
            )

    def as_source(self, as_stream=None) -> TableDataSource:
        """
        Generate a table data source with the same properties as the sink.

        Parameters
        ----------
        as_stream:
            If `True`, sink will be read as stream.

        Returns
        -------
        :
            Table Data Source
        """
        source = TableDataSource(
            catalog_name=self.catalog_name,
            table_name=self.table_name,
            schema_name=self.schema_name,
            warehouse=self.warehouse,
        )
        if as_stream:
            source.as_stream = as_stream

        return source



from typing import Literal
from laktory.models.basemodel import BaseModel


class ExternalLocationGrant(BaseModel):
    """
    Privileges granted to a principal and operating on an external location

    Attributes
    ----------
    principal
        User, group or service principal name
    privileges
        List of allowed privileges

    References
    ----------
    * [privilege types](https://docs.databricks.com/en/sql/language-manual/sql-ref-privileges.html#privilege-types)
    """

    principal: str
    privileges: list[
        Literal[
            "ALL_PRIVILEGES",
            "CREATE_EXTERNAL_TABLE",
            "CREATE_EXTERNAL_VOLUME",
            "READ_FILES",
            "WRITE_FILES",
            "CREATE_MANAGED_STORAGE",
        ]
    ]



from typing import Union
from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource
from laktory.models.resources.databricks.serviceprincipalrole import (
    ServicePrincipalRole,
)
from laktory.models.resources.databricks.groupmember import GroupMember


class ServicePrincipal(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks account service principal

    Attributes
    ----------
    allow_cluster_create:
        When `True`, the group is allowed to have cluster create permissions
    application_id:
        This is the Azure Application ID of the given Azure service principal
        and will be their form of access and identity. On other clouds than
        Azure this value is auto-generated.
    disable_as_user_deletion:
        If `True` user is disabled instead of delete when the resource is deleted
    display_name:
        Display name for the service principal
    group_ids:
        List of the group ids that the user should be member of.
    roles:
        List of roles assigned to the user e.g. ("account_admin")

    Examples
    --------
    ```py
    from laktory import models

    sp = models.resources.databricks.ServicePrincipal(
        display_name="neptune",
        application_id="baf147d1-a856-4de0-a570-8a56dbd7e234",
        group_ids=[
            "${resources.group-role-engineer.id}",
            "${resources.group-role-analyst.id}",
            "${resources.group-domain-finance.id}",
            "${resources.group-domain-engineering.id}",
        ],
        roles=["account_admin"],
    )
    ```
    """

    allow_cluster_create: bool = False
    application_id: str = None
    disable_as_user_deletion: bool = False
    display_name: str
    group_ids: list[str] = []
    roles: list[str] = []

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:ServicePrincipal"

    @property
    def resource_key(self) -> str:
        return self.display_name

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - service principal roles
        - service principal group members
        """
        resources = []
        for role in self.roles:
            resources += [
                ServicePrincipalRole(
                    service_principal_id=f"${{resources.{self.resource_name}.id}}",
                    role=role,
                )
            ]

        # Group Member
        for group_id in self.group_ids:
            resources += [
                GroupMember(
                    group_id=group_id,
                    member_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["groups", "roles", "group_ids"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_service_principal"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



spark_installed = True
try:
    import pyspark
except ModuleNotFoundError:
    spark_installed = False

if spark_installed:
    from pyspark.sql.connect.dataframe import DataFrame as SparkConnectDataFrame
    from pyspark.sql.dataframe import DataFrame as SparkDataFrame
    from pyspark.sql.column import Column as SparkColumn
    from pyspark.sql.session import SparkSession

    import laktory.spark.dataframe
    import laktory.spark.functions
    from laktory.spark.datatypes import DATATYPES_MAP

    def is_spark_dataframe(df):
        """Check if dataframe is Spark DataFrame or Spark Connect DataFrame"""

        if isinstance(df, SparkConnectDataFrame):
            return True

        if isinstance(df, SparkDataFrame):
            return True

        return False

else:
    # Mocks when pyspark is not installed

    class SparkDataFrame:
        pass

    class SparkConnectDataFrame:
        pass

    class SparkColumn:
        pass

    class SparkSession:
        pass

    def is_spark_dataframe(df):
        return True



import os.path

from typing import Literal
from typing import Any
from pydantic import model_validator

from laktory.models.datasources.basedatasource import BaseDataSource
from laktory.spark import SparkDataFrame
from laktory.polars import PolarsDataFrame
from laktory._logger import get_logger

logger = get_logger(__name__)


class FileDataSource(BaseDataSource):
    """
    Data source using disk files, such as data events (json/csv) and
    dataframe parquets. It is generally used in the context of a data pipeline.

    Attributes
    ----------
    format:
        Format of the data files
    header
        If `True`, first line of CSV files is assumed to be the column names.
    multiline
        If `True`, JSON files are parsed assuming that an object maybe be
        defined on multiple lines (as opposed to having a single object
        per line)
    read_options:
        Other options passed to `spark.read.options`
    schema_location:
        Path for files schema. If `None`, parent directory of `path` is used

    Examples
    ---------
    ```python
    from laktory import models

    source = models.FileDataSource(
        path="/Volumes/sources/landing/events/yahoo-finance/stock_price",
        format="JSON",
        as_stream=False,
    )
    # df = source.read(spark)
    ```
    """

    format: Literal["CSV", "PARQUET", "DELTA", "JSON", "EXCEL"] = "JSON"
    header: bool = True
    multiline: bool = False
    path: str
    read_options: dict[str, str] = {}
    schema_location: str = None

    @model_validator(mode="after")
    def options(self) -> Any:

        if self.dftype == "SPARK":
            if self.format in [
                "EXCEL",
            ]:
                raise ValueError(f"'{self.format}' format is not supported with Spark")

        return self

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def _id(self):
        return str(self.path)

    # ----------------------------------------------------------------------- #
    # Readers                                                                 #
    # ----------------------------------------------------------------------- #

    def _read_spark(self, spark) -> SparkDataFrame:
        if self.as_stream:
            logger.info(f"Reading {self._id} as stream")

            schema_location = self.schema_location
            if schema_location is None:
                schema_location = os.path.dirname(self.path)

            # Set reader
            reader = (
                spark.readStream.format("cloudFiles")
                .option("cloudFiles.format", self.format)
                .option("cloudFiles.schemaLocation", schema_location)
                .option("cloudFiles.inferColumnTypes", True)
                .option("cloudFiles.schemaEvolutionMode", "addNewColumns")
                .option("cloudFiles.allowOverwrites", True)
            )

        else:
            logger.info(f"Reading {self._id} as static")

            # Set reader
            reader = spark.read.format(self.format)

        reader = (
            reader.option("multiLine", self.multiline)  # only apply to JSON format
            .option("mergeSchema", True)
            .option("recursiveFileLookup", True)
            .option("header", self.header)  # only apply to CSV format
        )
        if self.read_options:
            reader = reader.options(**self.read_options)

        # Load
        df = reader.load(self.path)

        # Not supported by UC
        # .withColumn("file", F.input_file_name())

        return df

    def _read_polars(self) -> PolarsDataFrame:

        import polars as pl

        if self.as_stream:
            raise ValueError(
                "Streaming read not supported with Pandas DataFrame. Please switch to Spark"
            )

        logger.info(f"Reading {self._id} as static")

        if self.format.lower() == "csv":
            df = pl.read_csv(self.path, **self.read_options)

        elif self.format.lower() == "delta":
            df = pl.read_delta(self.path, **self.read_options)

        elif self.format.lower() == "excel":
            df = pl.read_excel(self.path, **self.read_options)

        elif self.format.lower() == "json":
            if self.multiline:
                df = pl.read_ndjson(self.path, **self.read_options)
            else:
                df = pl.read_json(self.path, **self.read_options)

        elif self.format.lower() == "parquet":
            df = pl.read_parquet(self.path, **self.read_options)

        else:
            raise ValueError(f"Format '{self.format}' is not supported.")

        return df



from pydantic import field_validator
from pydantic import model_validator
from typing import Any
from typing import Union
from typing import Callable
import abc

from laktory._logger import get_logger
from laktory.constants import SUPPORTED_DATATYPES
from laktory.models.basemodel import BaseModel
from laktory.polars import PolarsDataFrame
from laktory.polars import PolarsExpr
from laktory.types import AnyDataFrame

logger = get_logger(__name__)


# --------------------------------------------------------------------------- #
# Helper Classes                                                              #
# --------------------------------------------------------------------------- #


class BaseChainNodeFuncArg(BaseModel):
    value: Union[Any]

    @field_validator("value")
    def value_to_data_source(cls, v: Any) -> Any:
        """
        Data source can't be set as an expected value type as it would create
        a circular dependency. Instead, we check at validation if the value
        could be instantiated as a DataSource object.
        """
        from laktory.models.datasources import classes

        for c in classes:
            try:
                v = c(**v)
                break
            except:
                pass

        return v

    @abc.abstractmethod
    def eval(self):
        raise NotImplementedError()

    def signature(self):
        from laktory.models.datasources import PipelineNodeDataSource
        from laktory.models.datasources import FileDataSource
        from laktory.models.datasources import TableDataSource
        if isinstance(self.value, PipelineNodeDataSource):
            return f"node.{self.value.node_name}"
        elif isinstance(self.value, FileDataSource):
            return f"file {self.value.path}"
        elif isinstance(self.value, TableDataSource):
            return f"table {self.value.full_name}"
        return str(self.value)


class BaseChainNodeColumn(BaseModel):
    name: str
    type: str = "string"
    unit: Union[str, None] = None
    expr: Union[str, None] = None
    sql_expr: Union[str, None] = None

    @model_validator(mode="after")
    def expression_valid(self) -> Any:
        if not (self.expr or self.sql_expr):
            raise ValueError("Either `expr` or `sql_expr` must be defined.")
        return self

    @field_validator("type")
    def check_type(cls, v: str) -> str:
        if "<" in v:
            return v
        else:
            if v not in SUPPORTED_DATATYPES:
                raise ValueError(
                    f"Type {v} is not supported. Select one of {SUPPORTED_DATATYPES}"
                )
        return v

    @abc.abstractmethod
    def eval(self, udfs=None):
        raise NotImplementedError()

        # Adding udfs to global variables
        if udfs is None:
            udfs = {}
        for k, v in udfs.items():
            globals()[k] = v

        if self.dataframe_type == "SPARK":

            # Imports required to evaluate expressions
            import pyspark.sql.functions as F
            from pyspark.sql.functions import col
            from pyspark.sql.functions import lit

            if self.sql_expr:
                return F.expr(self.sql_expr)

        elif self.dataframe_type == "POLARS":

            # Imports required to evaluate expressions
            import polars as pl
            import polars.functions as F
            from polars import col
            from polars import lit

            if self.sql_expr:
                return pl.Expr.laktory.sql_expr(self.sql_expr)

        expr = eval(self.expr)

        # Cleaning up global variables
        for k, v in udfs.items():
            del globals()[k]

        return expr


# --------------------------------------------------------------------------- #
# Main Class                                                                  #
# --------------------------------------------------------------------------- #


class BaseChainNode(BaseModel):

    func_args: list[Union[Any]] = []
    func_kwargs: dict[str, Union[Any]] = {}
    func_name: Union[str, None] = None
    sql_expr: Union[str, None] = None
    with_column: Union[BaseChainNodeColumn, None] = None
    with_columns: Union[list[BaseChainNodeColumn], None] = []
    _parent: "BaseChain" = None
    _parsed_func_args: list = None
    _parsed_func_kwargs: dict = None

    @model_validator(mode="after")
    def selected_flow(self) -> Any:

        if len(self._with_columns) > 0:
            if self.func_name:
                raise ValueError(
                    "`func_name` should not be set when using `with_column`"
                )
            if self.sql_expr:
                raise ValueError(
                    "`sql_expr` should not be set when using `with_column`"
                )
        else:
            if self.func_name and self.sql_expr:
                raise ValueError(
                    "Only one of `func_name` and `sql_expr` should be set."
                )
            if not (self.func_name or self.sql_expr):
                raise ValueError(
                    "Either `func_name`, `sql_expr` or `with_column` should be set when using"
                )

        return self

    @property
    def _with_columns(self) -> list[BaseChainNodeColumn]:
        with_columns = [c for c in self.with_columns]
        if self.with_column:
            with_columns += [self.with_column]
        return with_columns

    @property
    def is_column(self):
        return len(self._with_columns) > 0

    @property
    def id(self):
        if self.is_column:
            return "-".join([c.name for c in self.with_columns])
        return "df"

    # ----------------------------------------------------------------------- #
    # Class Methods                                                           #
    # ----------------------------------------------------------------------- #

    @abc.abstractmethod
    def execute(
        self,
        df: AnyDataFrame,
        udfs: list[Callable[[...], Union[PolarsExpr, PolarsDataFrame]]] = None,
        return_col: bool = False,
    ) -> Union[AnyDataFrame]:
        raise NotImplementedError()



import os
import pandas as pd
from pyspark.sql import SparkSession
import polars as pl

spark = SparkSession.builder.appName("UnitTesting").getOrCreate()
spark.conf.set("spark.sql.session.timeZone", "UTC")


# --------------------------------------------------------------------------- #
# DataFrames                                                                  #
# --------------------------------------------------------------------------- #

data_dirpath = os.path.join(os.path.dirname(__file__), "../../tests/data/")

# Spark
df_brz = spark.read.parquet(os.path.join(data_dirpath, "brz_stock_prices"))
df_slv = spark.read.parquet(os.path.join(data_dirpath, "slv_stock_prices"))
df_meta = spark.read.parquet(os.path.join(data_dirpath, "slv_stock_meta"))
df_name = spark.createDataFrame(
    pd.DataFrame(
        {
            "symbol3": ["AAPL", "GOOGL", "AMZN"],
            "name": ["Apple", "Google", "Amazon"],
        }
    )
)

# Polars
dirpath = os.path.join(data_dirpath, "slv_stock_prices")
for filename in os.listdir(dirpath):
    if filename.endswith(".parquet"):
        break
df_slv_polars = pl.read_parquet(os.path.join(dirpath, filename))

dirpath = os.path.join(data_dirpath, "slv_stock_meta")
for filename in os.listdir(dirpath):
    if filename.endswith(".parquet"):
        break
df_meta_polars = pl.read_parquet(os.path.join(dirpath, filename))



from pydantic import Field
from pydantic import AliasChoices
from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class Grant(BaseModel):
    """
    Grants grant

    Attributes
    ----------
    principal:
        User, group or service principal name
    privileges:
        List of allowed privileges
    """

    principal: str
    privileges: list[str]


class Grants(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Grants

    List of grants assigned to a securable object.

    Attributes
    ----------
    grants:
        List of grant assigned to the selected object
    catalog:
        Name of the catalog to assign the grants to
    external_location:
        Name of the external location to assign the grants to
    metastore:
        Name of the metastore to assign the grants to
    model
        Name of the user to assign the permission to.
    schema:
        Name of the schema to assign the permission to.
    share:
        Name of the share to assign the permission to.
    storage_credential:
        Name of the storage credential to assign the permission to.
    view:
        Name of the view to assign the permission to.
    volume:
        Name of the volume to assign the permission to.

    Examples
    --------
    ```py
    from laktory import models

    grants = models.resources.databricks.Grants(
        catalog="dev",
        grants=[{"principal": "metastore-admins", "privileges": ["CREATE_SCHEMA"]}],
    )
    ```
    """

    grants: list[Grant]
    catalog: str = None
    external_location: str = None
    metastore: str = None
    model: str = None
    schema_: str = Field(
        None, validation_alias=AliasChoices("schema", "schema_")
    )  # required not to overwrite BaseModel attribute
    share: str = None
    storage_credential: str = None
    view: str = None
    volume: str = None

    # ----------------------------------------------------------------------- #
    # Pulumi Methods                                                          #
    # ----------------------------------------------------------------------- #
    @property
    def pulumi_renames(self) -> dict[str, str]:
        return {"schema_": "schema"}

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Grants"

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_grants"

    @property
    def terraform_renames(self) -> dict[str, str]:
        return self.pulumi_renames
